/////////////// code generation routines for NASM //////////////

/**************************************************************/
/* all the CPU and assembler dependend code is collected here */
/* you can port compiler to another processor simply by       */
/* modifying the following functions. It's a snap!            */
/*                                                            */
/* this implemention is for 386+ under Linux with NASM        */
/*                                                            */
/* convention: the name of the function is of the type        */
/* OpcodeDestSource with some exceptions. To clean up syntax  */
/* a bit Dest is omitted if it is the accumulator (LoadAccVar */
/* becomes simply LoadVar, SetFalse means SetAccFalse) and    */
/* "And" is omitted (e.g. PopAndMulAcc will be written simply */
/* as PopMul). Other abbreviations are:                       */
/*                                                            */
/* SetIfEqual -> SetEqual                                     */
/* SetIfNotEqual -> SetNEqual                                 */
/*                                                            */
/* i could also use macro with #define instead                */
/* of inline functions                                        */
/**************************************************************/

#include "main.h"
#include "scanner.h"


/* Load a Variable to Primary Register */
inline void LoadVar(string Name)
{
   /* load a variabile of name Name into accumulator */
   EmitLn("MOV EAX,[" + Name + "]");
}

/* calls a procedure called Name */
inline void CallProc(string Name)
{
   EmitLn("CALL " + Name);
}

/* Load a Constant Value to Primary Register */
inline void LoadConst(string Name)
{
   /* loads a constant value into accumulator */
   EmitLn("MOV EAX," + Name);
}

/* loads a negative constant value into accumulator */
inline void LoadNegConst(string Name)
{
   EmitLn("MOV EAX,-" + Name);
}

/* Multiply Top of Stack by Primary */
inline void PopMul()
{
  /* pops a value from the system stack,
     multiplies it by accumulator and
     stores the result into accumulator. */


   /* this way uses EBX as a temporary */
   EmitLn("POP EBX");
   EmitLn("IMUL EBX");  // result is in EDX:EAX, EDX ignored

   /* another way not using EBX */
   /*
   EmitLn("IMUL DWORD [ESP]");
   EmitLn("ADD ESP,4");
   */
}

/* Divide Top of Stack by Primary */
inline void PopDiv()
{
   /* divides the operand on top of stack by the accumulator
      and leaves the result into accumulator */


/* from Intel's documentation:

CDQ - Convert Double to Quad (386+)

        Usage:  CDQ
        Modifies flags: None

        Converts signed DWORD in EAX to a signed quad word in EDX:EAX by
        extending the high order bit of EAX throughout EDX
*/

   EmitLn("MOV EBX,EAX");
   EmitLn("POP EAX");
   // dividend is EDX:EAX, so we can't let an uninitialized value in EDX
   // we can't zero out EDX with a XOR EDX,EDX, because this is a signed
   // division. we need to extend EAX sign through EDX. CDQ does this.
   EmitLn("CDQ");
   EmitLn("IDIV EBX"); // result is in EAX; EDX is remainder
}

inline void PopMod()
{
   EmitLn("MOV EBX,EAX");
   EmitLn("POP EAX");
   EmitLn("CDQ");
   EmitLn("IDIV EBX");
   EmitLn("MOV EAX,EDX");
}

/* Push Primary onto Stack */
inline void Push()
{
   /* pushs the accumulator into the stack */
   EmitLn("PUSH EAX");
}

/* pops the accumulator from the stack */
inline void Pop()
{
   EmitLn("POP EAX");
}

/* Add Top of Stack to Primary */
inline void PopAdd()
{
  /* pops the stack and add the popped value to accumulator
     leaves the result into the accumulator */
   EmitLn("POP EBX");
   EmitLn("ADD EAX,EBX");
}

/* Subtract Primary from Top of Stack */
inline void PopSub()
{
   /* pop a value from the stack and subtracts accumulator from it
      leaves the result into the accumulator */
   EmitLn("POP EBX");
   EmitLn("SUB EAX,EBX");
   EmitLn("NEG EAX");
}

/* Clear the Primary Register */
inline void Clear()
{
   /* zeroes out the accumulator */
   EmitLn("XOR EAX,EAX");
}

/* Negate the Primary Register */
inline void Negate()
{
   /* two's complement negation of accumulator */
   /* (change the accumulator sign). */
   EmitLn("NEG EAX");
}

/* zeroes out the accumulator, or better sets it to
   the logical value False without altering the
   condition flags (in that is different from ClearAcc()) */
inline void SetFalse()
{
   // XOR EAX,EAX is faster, but alters the cc */
   EmitLn("MOV EAX,0");
}

/* performs a bit-a-bit not of accumulator */
inline void BitABitNot()
{
   EmitLn("NOT EAX");
}

/* decrements the accumulator */
inline void Dec()
{
   EmitLn("DEC EAX");
}

/* Store Primary to Variable */
inline void StoreVar(string Name)
{
   /* stores the accumulator into the specified variable */
   EmitLn("MOV [" + Name + "],EAX");
}

/* note: NEAR is necessary because the ordinary form of the instructions
   has only a 128-byte range in NASM; the NEAR form is a 386 extension to
   the instruction set, and can span the full size of a segment.
   The simplest thing is to add NEAR to all conditional jump instruction
   whether the jump is 128 byte long or not. */
   
/* Branch False */
inline void BranchFalse(string Label)
{
   /* branch if false
      (accumulator is zero or last operation had a result of zero
       or involved a zero; this is machine-dependant) */
   /* condition flags has to be properly set before call this */
   EmitLn("JZ NEAR " + Label);
}

/* this is the opposite of the previous */
inline void BranchTrue(string Label)
{
   EmitLn("JNZ NEAR " + Label);
}

/* Branch Unconditional */
inline void Branch(string Label)
{
/* unconditional jump */
   EmitLn("JMP " + Label);
}

/* jump to label if greater than */
/* you must compare before calling this
   (condition flags have to be setted the right way) */
inline void BranchGreater(string Label)
{
   EmitLn("JG NEAR " + Label);
}

inline void BranchLessOrEqual(string Label)
{
   EmitLn("JLE NEAR " + Label);
}

inline void BranchLess(string Label)
{
   EmitLn("JL NEAR " + Label);
}

inline void BranchGreaterOrEqual(string Label)
{
   EmitLn("JGE NEAR " + Label);
}

/* discards the current value (of double word size) on the top of stack */
inline void DiscardDWORDFromStack()
{
   EmitLn("ADD ESP,4");
}

/* loads the top of stack into accumulator, without popping it out */
inline void LoadFromStack()
{
   EmitLn("MOV EAX,[ESP]");
}

/* increments the given variable */
inline void IncVar(string Name)
{
   EmitLn("INC DWORD[" + Name + "]");
}

/* compares the given variabile with the accumulator
   and sets the condition flags accordingly */
inline void CompareVar(string Name)
{
   EmitLn("CMP [" + Name + "], EAX");
}

/* compares the value of the accumulator with zero */
inline void CompareZero()
{
   EmitLn("OR EAX,EAX"); // as CMP EAX,0 but a little faster
}

/* loads the Boolean value FALSE representation into accumulator */
/* this is currently a zero */
inline void LoadFALSE()
{
   EmitLn("XOR EAX,EAX");
}

/* as above but loades the Boolean value TRUE */
/* i've chosen -1 */
inline void LoadTRUE()
{
   EmitLn("MOV EAX,-1");
}

/* OR Top of Stack with Primary */
inline void PopOr()
{
   /* performs a bit-a-bit or between accumulator and the top of stack */
   /* then pops the stack */
   EmitLn("OR EAX,[ESP]");
   EmitLn("ADD ESP,4");
}

/* XOR Top of Stack with Primary */
inline void PopXor()
{
   /* as above but operation is a xor (esclusive or) */
   EmitLn("XOR EAX,[ESP]");
   EmitLn("ADD ESP,4");
}

/* AND Top of Stack with Primary */
inline void PopAnd()
{
   /* this function pops a value from the stack and
      performs an And bit-a-bit between this value
      and the accumulator. The result is left into
      accumulator */
   EmitLn("AND EAX,[ESP]");
   EmitLn("ADD ESP,4");
}


/* inverts the logical value of accumulator */
/* and sets the condition flags accordingly */
inline void Not()
{
   EmitLn("XOR EAX,-1");
   // NOT EAX would be optmized,
   // but isn't good because it doesn't affect the condition flags
}

/* sets the condition flags (mainly the zero flag)
   according to the logical value of accumulator */
inline void Test()
{
   EmitLn("TEST EAX,EAX"); // OR EAX,EAX or AND EAX,EAX is also good
}

/* Compare Top of Stack with Primary */
inline void PopCompare()
{
   /* does what the name says: pops a value from the stack
      and compares it with accumulator (sets the condition flags).
      note the trick to let the condition flags with the right value */
   EmitLn("POP EBX");
   EmitLn("CMP EBX,EAX");

   // this is wrong: the ADD alters the cc
   //   EmitLn("CMP [ESP], EAX");
   //   EmitLn("ADD ESP,4");

   // other right solution is (not using EBX as a temporary):
   /*EmitLn("ADD ESP,4");
   EmitLn("CMP [ESP-4], EAX");*/
}

/** these are used by the boolean parser
    to emit code for relational operators.

    a bit of complexity arises, because the
    386+ has not a suitable Scc instruction
    as the 68000 - SETcc is not good,
    because it can only set a byte to 0 and 1,
    not a dword to 0 or FFFFFFFFh as required
    by this 32 bit implementation of KISS **/

/* Set Primary If Compare was = */
inline void SetEqual()
{
   /* --implementation notes:
      note that instead of XOR EAX,EAX, we use a MOV EAX,0
      which is a little slower, but we need to preserve the
      condition flags. Also we use NOT EAX *instead of*
      Not() (we have made another code generation
      function for NOT EAX). This is another little optimization.
      Remember: this initial version of the KISS compiler isn't
      optimized optimized at all, but this little goodies doesn't
      cost us much. */

   string L;

   SetFalse();                  // CODE emission

   L = NewLabel();

   BranchTrue(L);               // CODE emission

   /* this makes EAX=-1 (which is TRUE) */
   BitABitNot();                // CODE emission

   PostLabel(L);
}

/* Set Primary If Compare was != */
inline void SetNEqual()
{
   string L;

   SetFalse();                  // CODE emission

   L = NewLabel();

   BranchFalse(L);              // CODE emission

   /* this makes EAX=-1 (which is TRUE) */

   BitABitNot();                // CODE emission

   PostLabel(L);
}

/* Set Primary If Compare was < */
inline void SetLess()
{
   string L;

   SetFalse();                     // CODE emission

   L = NewLabel();

   BranchGreaterOrEqual(L);        // CODE emission

   /* this makes EAX=-1 (which is TRUE) */

   BitABitNot();                   // CODE emission

   PostLabel(L);
}

/* Set Primary If Compare was > */
inline void SetGreater()
{
   string L;

   SetFalse();               // CODE emission

   L = NewLabel();

   BranchLessOrEqual(L);     // CODE emission

   BitABitNot();             // CODE emission

   PostLabel(L);
}

/* Set Primary If Compare was <= */
inline void SetLessOrEqual()
{
   string L;

   SetFalse();                     // CODE emission

   L = NewLabel();

   BranchGreater(L);               // CODE emission

   /* this makes EAX=-1 (which is TRUE) */

   BitABitNot();                   // CODE emission

   PostLabel(L);
}

/* Set Primary If Compare was >= */
inline void SetGreaterOrEqual()
{
   string L;

   SetFalse();               // CODE emission

   L = NewLabel();

   BranchLess(L);            // CODE emission

   BitABitNot();             // CODE emission

   PostLabel(L);
}


/**********************************************************/
/* directives generation: this is much assembler specific */
/**********************************************************/

/*------------------------------------------------------------*/
/* Write Header Info */

inline void Header()
{
   EmitLn(";; generated by Tiny KISS 0.1 - distributed under GPL");
   EmitLn(";; by Antonio Bonifati <abonifati@tiscalinet.it>\n");
   EmitLn("bits 32\n");
   EmitLn("section .data");
   EmitLn("_readint: db \"%d\",0");
   EmitLn("_writeint: db \"%d\",10,0\n");
}


/*------------------------------------------------------------*/
/* Write the Prolog */

inline void Prolog()
{
   cout << endl;
   EmitLn("section .text");
   EmitLn("extern scanf");
   EmitLn("extern printf");
   EmitLn("global main\n");
   PostLabel("main");
}

/*------------------------------------------------------------*/
/* Write the Epilog */

inline void Epilog()
{
   cout << endl;
   EmitLn("MOV EAX,1");         // eax=1, _exit () system call
   EmitLn("XOR EBX,EBX");       // ebx=0, returned to OS
   EmitLn("INT 0x80");          // execute system call
}

/*------------------------------------------------------------*/
/* Allocate (initialized) Storage for a Variable              */

inline void ReserveVar(string N)
{
   cout << N << ':' << TAB << "DD ";

   if (Look == '=')
   {
      Match('=');
      if (Look == '-')
      {
         cout << Look;
         Match('-');
      }
      GetNum();
      cout << Value << endl;
   }
   else
      cout << "0\n";
}


/************************************************************/
/* i/o routines, or interface to some external i/o routines */
/* could something be more os-dependant than this? :)       */
/************************************************************/

/*------------------------------------------------------------*/
/* Read Variable to Primary Register */

inline void ReadVar(string Name)
{
   /* this is obtained by accessing libc functions via assembler    */
   /* DWORD is not optional; scanf must be written in all lowercase */
   /* parameters must be pushed from RIGHT to LEFT, not from LEFT
      to RIGHT. This C convention allows to have a variable number
      of parameters. */
   EmitLn("PUSH DWORD " + Name);  // push variable address
   EmitLn("PUSH DWORD _readint"); // push format string ("%d")
   EmitLn("CALL scanf");          // call libc function
   EmitLn("ADD ESP,8");           // scanf doesn't pop parameters from
                                  // the stack
                                  // here BYTE before 8 is optional

   // not required with scanf
   //StoreVar(Value);
}


/*------------------------------------------------------------*/
/* Write Variable from Primary Register */

inline void WriteVar()
{
   EmitLn("PUSH EAX");
   EmitLn("PUSH DWORD _writeint");
   EmitLn("CALL printf");
   EmitLn("ADD ESP,8");         // clean stack
}

