// KISS Compiler Version 0.1
// Distributed under GPL
// Author: Antonio Bonifati

// Notes: this is the program parser (the main parser)
// it also handles variables and procedures/functions
// declarations, the symbol table, the bult-in
// I/O statements.

/*------------------------------------------------------------*/
                          /* Prog */


#include "prog.h"
#include "expr.h"
#include "stat.h"
#include "error.h"
#include "scanner.h"
#include "codegen.h"


const int MaxEntry = 100;
int NEntry=0;

Symbol ST[MaxEntry];        /* Simbol Table        */


/*------------------------------------------------------------*/
/* Look for Symbol in Table */
// should be inline
int InTable(string n)
{
   return(Lookup(ST, n, NEntry));
}


/*------------------------------------------------------------*/
/* Add a New Entry to Symbol Table */

/* uses insertion sort to keep the Symbol Table in order
   and allow fast search (binary search) on it. */
void AddEntry(Symbol N)
{
   //if (InTable(N) != NOTFOUND) Abort("Duplicate Identifier " + N);
   if (NEntry++ == MaxEntry) Abort("Symbol Table Full");
   for (int i=NEntry; i>low; i--)
      ST[i] = ST[i-1];
   ST[low] = N;       /* or i */
}


/*------------------------------------------------------------*/
/* Process a Read Statement */
void Read()
{
   Match('(');
   GetName();
   ReadVar(Value);
   NewLine();
   while (Look == ',')
   {
      Match(',');
      GetName();
      ReadVar(Value);
      NewLine();
   }
   Match(')');
}


/*------------------------------------------------------------*/
/* Process a Write Statement */

void Write()
{
   Match('(');
   Expression();
   WriteVar();
   NewLine();
   while (Look == ',')
   {
      Match(',');
      Expression();
      WriteVar();
      NewLine();
   }
   Match(')');
}


/*------------------------------------------------------------*/
/* Recognize and Translate a Statement Block                  */
/*                                                            */
/* the exit address of the innermost loop is passed to block  */
/*                                                            */
/* <block> ::= [<statement>]*                                 */
/* <statement> ::= <if> | <while> | <assignment> ...          */

void block(string L)
{
   Scan();
   while (!(Token=='e'  ||  Token=='l'  ||  Token=='u'))
   {
      switch(Token)
      {
         case 'i': If(L);       break;
         case 'w': While();     break;
         case 'p': Loop();      break;
         case 'r': Repeat();    break;
         case 'f': For();       break;
         case 'd': Do();        break;
         case 'k': Break(L);    break;
         case 'R': Read();      break;
         case 'W': Write();     break;
         default: Assignment();
      }
      Scan();
   }
}
/*------------------------------------------------------------*/


/*------------------------------------------------------------*/
/* Parse and Translate a Main Program                         */
/*                                                            */
/* <main> ::= BEGIN <block> END                               */

void Main()
{
   MatchString("BEGIN");
   
   Prolog();       // CODE generation

   block("");
   MatchString("END");
   
   Epilog();       // CODE generation
}


/*------------------------------------------------------------*/
/* Allocate Storage for a Variable */

void Alloc(string N)
{
   if (InTable(N)!=NOTFOUND) Abort("Duplicate Variable Name " + N);
   AddEntry(N);
   
   ReserveVar(N);  // CODE generation
}


/*------------------------------------------------------------*/
/* Process a Data Declaration                                 */
/*                                                            */
/* <data declaration> ::= VAR <var-list>                      */
/* <var-list> ::= <ident> (, <ident>)*                        */

void Decl()
{
   GetName();
   Alloc(Value);
   NewLine();
   while (Look == ',')
   {
      Match(',');       // GetChar() doesn't skip whites.
      GetName();
      Alloc(Value);
      NewLine();
   }
}


/*------------------------------------------------------------*/
/* Parse and Translate Global Declarations                    */
/*                                                            */
/* <top-level decls> ::= ( <data declaration> )*              */

void TopDecls()
{
   Scan();
   while (Token != 'b') // b is the code for BEGIN
   {
      switch(Token)
      {
         case 'v': Decl();      break;
         default: Abort("Unrecognized Keyword '" + Value + "'");
      }
      Scan();
   }
}


/*------------------------------------------------------------*/
/* Parse and Translate a Program                              */
/*                                                            */
/* <program> ::= PROGRAM <top-level decl> <main> '.'          */

void Program()
{
   MatchString("PROGRAM");      // don't OPTIMIZED (there's not a code for
                                // PROGRAM), but it doesn't matter much
                                // because it always appears once in a program
                                
   Header();       // CODE generation

   TopDecls();
   Main();
   Match('.');
}
/*------------------------------------------------------------*/

/*------------------------------------------------------------*/
/* Initialize */

void InitProg()
{
// not necessary
//   for (int i=0; i<MaxEntry; i++)
//      ST[i] = "";
}

