// KISS Compiler Version 0.1
// Distributed under GPL
// Author: Antonio Bonifati

// Notes: the lexical scanner (this is of type distributed)
// this part is very efficient
// the scanner takes also care of comments when feasible


/*------------------------------------------------------------*/
                          /* Scanner */


#include <iostream>
#include "scanner.h"
#include "recogn.h"
#include "error.h"


/*------------------------------------------------------------*/
/* Definition of Keywords and Token Types                     */
/* this list MUST be ordered,
   'cause Lookup uses binary searching */

/* Because in the parser code (stat.cpp) there are calls like
                         MatchString("ENDWHILE");
                         MatchString("UNTIL");
                         ...
   for that keywords that maps on the same code, if you change some
   keyword names, you'll need to change MatchString argument accordingly.
   Don't forget this! */
Symbol KWlist[] = { "BEGIN", "BREAK", "DO", "ELSE", "END", "ENDDO", "ENDFOR",
                    "ENDIF", "ENDLOOP", "ENDWHILE", "FOR", "IF", "LOOP",
                    "READ", "REPEAT", "TO", "UNTIL", "VAR", "WHILE", "WRITE" };

/* don't modify this */
#define KWnumber (sizeof KWlist / sizeof(Symbol))
/* another way: */
//#define KWnumber (sizeof KWlist / sizeof KWlist[0])

/* this must correspond to the previous array. */
/* remember that 'x' is used for ident and '#' for numbers
   and '?' for all other things. */
/* you can make two or more keywords have the same code,
   if you make the parser take care of distinguish them.
   Note that to speed up things you could avoid this if
   the keywords beginning with the same letter are few;
   for example here "loop" is encoded with p because
   l was chosen to encode "else" yet */
const string KWcode = "bkdleeeeeefipRrtuvwW"; /* i'm not mad ;) */
/*------------------------------------------------------------*/


/*------------------------------------------------------------*/
/* Variable Declarations */

char Look;                    /* Lookahead Character */
char Token;                   /* Current Encoded Token */
string Value;                 /* String Token of Look (Scanned or Unencoded Token) */


/*------------------------------------------------------------*/
/* Table Lookup */

/* If the input string matches a table entry, return the entry
   index.  If not, returns NOTFOUND (-1) - defined in scanner.h */

/* uses a optimized binary search to speed up things.
   Please care to maintain TabPtr ordered when you add new keywords.

   This is also used to search the symbol table. Note that
   low (or high) is the insertion position. */
// should be optimized with only one test
int low;
int Lookup(TabPtr T, string s, int n)
{
   int  high, mid;

   low = 0;
   high = n - 1;

   while (low <= high)
   {
      mid = (low+high) / 2;
      if (s < T[mid])
          high = mid - 1;
      else  if (s > T[mid])
          low = mid + 1;
      else
          return mid;   /* found */
   }
   return(NOTFOUND);
}
/*------------------------------------------------------------*/


/*------------------------------------------------------------*/
/* Read New Character From Input Stream */

void GetChar(void)
{
   cin.get(Look);
}


/*------------------------------------------------------------*/
/* Skip Over Leading White Space (not a NEWLINE) */

void SkipWhite()
{
   while (IsWhite(Look))
      GetChar();
}


/*------------------------------------------------------------*/
/* Skip Over an End-of-Line */

void NewLine()
{
   while (IsLineSep(Look))
   {
      ++line;
      GetChar();
      SkipWhite();
   }
}


/*------------------------------------------------------------*/
/* skips a one line-comment or multiple one lines comments    */

void Comment()
{
   while (Look == '#')
   {
      do
      {
         GetChar();
      } while (!IsLineSep(Look));
      NewLine();
   }
}


/*------------------------------------------------------------*/
/* Match a Specific Input Character */
void Match(char x)
{
   NewLine();
   Comment();
   if (Look != x) Expected(string("'") + x + "'"); // or string("'") + x + '\''
   else
   {
      GetChar();
      SkipWhite();
   }

   /* or you can code: */
   /*NewLine();
   Comment();
   if (Look == x) GetChar();
   else Expected(string("'") + x + "'"); // or string("'") + x + '\''
   SkipWhite();*/
}

/*------------------------------------------------------------*/
/* Match a Specific Input String */

void MatchString(string x)
{
   if (Value != x) Expected("'" + x + "'");
}


/*------------------------------------------------------------*/
/* Get an Identifier                                          */
/*                                                            */
/* <name> ::= <letter> [<alnum>]*                             */

void GetName()
{
   NewLine();
   Comment();
   if (!isalpha(Look)) Expected("Name");
   Value = "";
   while (isalnum(Look))
   {
      Value += toupper(Look);
      GetChar();
   }
   SkipWhite();
}


/*------------------------------------------------------------*/
/* Get a Number                                               */
/*                                                            */
/* <number> ::= <digit> [<digit>]*                            */

void GetNum()
{
   NewLine();
   Comment();
   if (!isdigit(Look)) Expected("Integer");
   Value = "";
   while (isdigit(Look))
   {
      Value += Look;
      GetChar();
   }
   SkipWhite();

   Token='#'; /* code for a number */
}


/*------------------------------------------------------------*/
/* Get an Operator */
/*
void GetOp()
{
   Value = "";
   if (!IsOp(Look)) Expected("Operator");
   while (IsOp(Look))
   {
     Value += Look;
     GetChar();
   }
   if (Value.length() == 1)
      Token = Value[1];
   else
      Token = '?';
}
*/

/*------------------------------------------------------------*/
/* Get a Boolean Literal                                      */
/*                                                            */
/* <b-literal> ::= TRUE | FALSE                               */
// needed in KISS
//bool GetBoolean()
//{
//   char result;

//   if (!IsBoolean(Look)) Expected("Boolean Literal");
//   result = toupper(Look) == 'T';
//   GetChar();
//   return(result);
//}


/*------------------------------------------------------------*/
/* Lexical Scanner */

void Scan()
{
   int CODEpos;

   GetName();

   if ((CODEpos=Lookup(KWlist, Value, KWnumber)) == NOTFOUND)
      Token = 'x'; /* code for an ident */
   else
      Token = KWcode[CODEpos];
}


/*------------------------------------------------------------*/
/* Initialize */

void InitScanner()
{
   GetChar();
   SkipWhite();
   Scan();
}

