// KISS Compiler Version 0.1
// Distributed under GPL
// Author: Antonio Bonifati

// Notes: this is the statement parser (also handles assignments).


/*------------------------------------------------------------*/
                          /* Stat */


#include "stat.h"
#include "prog.h"
#include "scanner.h"
#include "codegen.h"
#include "boolexpr.h"
#include "error.h"


/*------------------------------------------------------------*/
/* Recognize and Translate an IF Construct                    */
/*                                                            */
/* L is the exit address (label) of the innermost loop        */
/*                                                            */
/* <if then else> ::= IF <condition> <block>                  */
/*                    [ELSE <block>] ENDIF                    */

void If(string L)
{
   string L1, L2;

   BoolExpression();
   L1 = NewLabel();
   L2 = L1;   
   
   BranchFalse(L1);      // CODE emission

   block(L);
   if (Token == 'l')
   {
      L2 = NewLabel();

      Branch(L2);        // CODE emission
      
      PostLabel(L1);
      block(L);
   }
   MatchString("ENDIF");
   PostLabel(L2);
}


/*------------------------------------------------------------*/
/* Parse and Translate a WHILE Statement                      */
/*                                                            */
/* <while> ::= WHILE <condition> <block> ENDWHILE             */

void While()
{
   string L1, L2;

   L1 = NewLabel();
   L2 = NewLabel();
   PostLabel(L1);
   BoolExpression();
   
   BranchFalse(L2);     // CODE emission

   block(L2);
   MatchString("ENDWHILE");

   Branch(L1);          // CODE emission
   
   PostLabel(L2);
}


/*------------------------------------------------------------*/
/* Parse and Translate a LOOP Statement                       */
/*                                                            */
/* <loop> ::= LOOP <block> ENDLOOP                            */

void Loop()
{
   string L1,L2;

   L1 = NewLabel();
   L2 = NewLabel();
   PostLabel(L1);
   block(L2);
   MatchString("ENDLOOP");

   Branch(L1);          // CODE emission

   PostLabel(L2);
}


/*------------------------------------------------------------*/
/* Parse and Translate a REPEAT Statement                     */
/*                                                            */
/* <repeat until> ::= REPEAT <block> UNTIL <condition>        */

void Repeat()
{
   string L1,L2; /* L2 is needed by a possible break */

   L1 = NewLabel();
   L2 = NewLabel();
   PostLabel(L1);
   block(L2);
   MatchString("UNTIL"); // here you can also test if Token=='u', which is faster
   BoolExpression();

   BranchFalse(L1);      // CODE emission

   PostLabel(L2);
}


/*----------------------------------------------------------------*/
/* Parse and Translate a FOR Statement                            */
/*                                                                */
/* <for loop> ::= FOR <ident> = <expr1> TO <expr2> <block> ENDFOR */

void For()
{
   string L1, L2;
   string Name;

   L1 = NewLabel();
   L2 = NewLabel();
   GetName();
   Name = Value;
   Match('=');
   BoolExpression();
   
   Dec();               // CODE emission

   if (InTable(Name)==NOTFOUND) Undefined(Name);
   
   StoreVar(Name);      // CODE emission

   Scan();
   MatchString("TO"); // see note beside MatchString("UNTIL");

   BoolExpression();

   Push();              // CODE emission
   
   PostLabel(L1);

   LoadFromStack();          // CODE emission
   IncVar(Name);             // CODE emission
   CompareVar(Name);         // CODE emission
   BranchGreater(L2);        // CODE emission

   block(L2);
   MatchString("ENDFOR");

   Branch(L1);          // CODE emission

   PostLabel(L2);

   DiscardDWORDFromStack(); // CODE emission
}


/*------------------------------------------------------------*/
/* Parse and Translate a DO Statement                         */
/*                                                            */
/* <do loop> ::= DO <expr> <block> ENDDO                      */
void Do()
{
// this was the original implementation by Jack Crenshaw

// string L1,L2;

   /* here are the syntax and translation chosen for the 386+ */
   /* i will use ECX in conjunction with the LOOP instruction  */
   /* i will save the value of ECX before calling block and
      then restore it after to avoid any future problem when
      the expression parser will be optimized (and will use ECX) */
   /*
     DO
     <expr>         { Emit(MOV ECX,EAX);
                      L1 = NewLabel;
                      PostLabel(L);
                      Emit(PUSH ECX) }
     <block>
     ENDDO          { Emit(POP ECX);
                      Emit(LOOP L1) }
   */
   
//   Match('d');
//   L1 = NewLabel();
//   L2 = NewLabel();
//   BoolExpression();

//   EmitLn("MOV ECX,EAX");
   
//   PostLabel(L1);

//   EmitLn("PUSH ECX");

//   block(L2);

//   EmitLn("POP ECX");
//   EmitLn("LOOP " + L1);

//   PostLabel(L2);


// this is the new implementation of the statement DO
//
// another way to code is to use the Intel LOOP instruction
// and ECX as counter. Because LOOP can be used for
// make loops with tail condition, you will also need
// to perform a head test to avoid to make an iteration if
// expression evaluates to zero (or negative).
   string L1, L2, L3;

   L1 = NewLabel();
   L2 = NewLabel();
   L3 = NewLabel();
   BoolExpression();
   PostLabel(L1);

   CompareZero();               // CODE emission
   BranchLessOrEqual(L3);       // CODE emission
   Push();                      // CODE emission

   block(L2);

   Pop();                       // CODE emission
   Dec();                       // CODE emission
   Branch(L1);                  // CODE emission

   MatchString("ENDDO"); // this can be optmized

   // PATCH to handle a break issued by block correctly
   // we have to ensure that stack is emproperly cleared
   // you can also use ADD ESP,4
   // instead of POP EAX
   // issuing this isn't optimized, but for the sake of
   // semplicity ...
   PostLabel(L2);

   Pop();                       // CODE emission

   PostLabel(L3);
}

/*------------------------------------------------------------*/
/* Recognize and Translate a BREAK                            */
/*                                                            */
/* a break jumps to the end of the innermost loop, which is   */
/* pointed by the label L passed. An attempt to exit the      */
/* outermost block is an error, so Program() passes a null    */
/* label which is caught by Break.                            */

void Break(string L)
{
   if (L != "")
      Branch(L);        // CODE emission
   else Abort("No loop to break from");
}


/*------------------------------------------------------------*/
/* Parse and Translate an Assignment Statement                */
/*                                                            */
/* <assignment> ::= <ident> = <b-expression>                  */

void Assignment()
{
   string Name;

   // ignore extra chars in names longer than IDMAXLEN
   Name.assign(Value, 0, IDMAXLEN);

   if (InTable(Name)==NOTFOUND) Undefined(Name);
   
   Match('=');
   BoolExpression();

   StoreVar(Name);      // CODE emission
}

